import Progress from '../util/Progress';
import { ClassMap, ClassMapEntity, Sources, Z8ClassProperties } from './Sources';
import fs from 'fs';
import { JSDOM, VirtualConsole } from 'jsdom';

class Analyzer {
	public static analyze(classMap: ClassMap | null, callback?: () => any, silent?: boolean): void {
		let script = fs.readFileSync(Sources.dzAppPath, 'utf8').toString();

		// replace Z8.$ function to prevent showing messages from Z8.locales in analyzed classes
		script = script.replace('return i18n.getMessage(key, data);', 'return `Z8.$(\'${key}\'${data ? \', ...\' : \'\'})`;');
		
		const virtualConsole = new VirtualConsole();
		virtualConsole.on('jsdomError', (error) => {});
		
		const dom = new JSDOM(`<!DOCTYPE html><html><head><meta http-equiv="Content-Type" content="text/html; charset=UTF-8"></head><body><script type="text/javascript">${script}</script></body></html>`, {
			virtualConsole: virtualConsole,
			runScripts: 'dangerously',
			resources: 'usable'
		});

		dom.window.document.addEventListener('DOMContentLoaded', () => {
			try {
				if(classMap) {
					const Z8Classes = (dom.window as any).Z8.classes;

					const progress = silent ? null : Progress.start('Sources update, Step 4 [Analyze]:', Object.keys(Z8Classes).length + 2 * Object.keys(classMap).length + 1);

					Analyzer.filterZ8Classes(Z8Classes, progress);
					Analyzer.analyzeZ8Classes(Z8Classes, classMap, progress);
					Analyzer.postProcessClassMap(classMap, progress);

					fs.writeFileSync(Sources.classMapPath, JSON.stringify(classMap), { flag: 'a+', encoding: 'utf8' });

					progress && progress.finish();
				}

				Sources.Z8Locales = (dom.window as any).Z8.locales;

				callback && callback();
			} catch (e) {
				console.log(e);
			}

			dom.window.close();
		});
	}

	private static filterZ8Classes(Z8Classes: any, progress: Progress | null): any {
		const classesToDelete = [];
		for(const key of Object.keys(Z8Classes)) {
			const shortClassName = Analyzer.getShortClassName(Z8Classes[key]);
			if(key !== shortClassName && shortClassName != null)
				classesToDelete.push(shortClassName);
			progress && progress.next();
		}

		progress && progress.sumTotal(classesToDelete.length);

		for(const cls of classesToDelete) {
			delete Z8Classes[cls];
			progress && progress.next();
		}
	}

	private static analyzeZ8Classes(Z8Classes: any, classMap: ClassMap, progress: Progress | null): void {
		for(const className of Object.keys(classMap)) {
			let cls = classMap[className];
			const z8Cls = Z8Classes[className];

			cls = Analyzer.setParentsBranch(cls, Analyzer.getParentsBranch(z8Cls, Z8Classes, classMap, [], null)!);
			cls = Analyzer.setExtends(cls, Analyzer.getSuperclassName(z8Cls));
			cls = Analyzer.setShortName(cls, Analyzer.getShortClassName(z8Cls));
			cls = Analyzer.setMixins(cls, Analyzer.getMixins(z8Cls, classMap));
			cls = Analyzer.setStatics(cls, Analyzer.getStatics(z8Cls));
			cls = Analyzer.setProperties(cls, Analyzer.getProperties(z8Cls));

			classMap[className] = cls;

			progress && progress.next();
		}
	}

	private static postProcessClassMap(classMap: ClassMap, progress: Progress | null): void {
		for(const className of Object.keys(classMap)) {
			const cls = classMap[className];
			const parents = cls.parentsBranch;
			const properties = cls.properties;
			const selfProperties = properties.map(item => item);
			let dynamicProperties = cls.dynamicProperties;

			for(let i = parents.length - 1; i >= 0; i--) {
				const parentName = parents[i];
				const parent = classMap[parentName];
				const parentProperties = parent.properties;
				const parentDynamicProperties = parent.dynamicProperties;

				for(const parentProperty of parentProperties) {
					const parentPropertyCopy = Object.assign({}, parentProperty);
					const propIndex = properties.findIndex((item, index) => item.key === parentProperty.key);

					parentPropertyCopy.inherited = true;
					parentPropertyCopy.nearestParent = parentName;
					parentPropertyCopy.nearestParentRoot = parent.root;
					parentPropertyCopy.overridden = false;

					if(propIndex === -1) {
						properties.push(parentPropertyCopy);
					} else {
						parentPropertyCopy.overridden = selfProperties.some((item) => item.key === parentPropertyCopy.key);
						if(parentPropertyCopy.overridden)
							parentPropertyCopy.value = selfProperties[propIndex].value;
						properties[propIndex] = parentPropertyCopy;
					}

					dynamicProperties = dynamicProperties.filter((item) => item.key !== parentProperty.key);
				}

				for(const parentProperty of parentDynamicProperties) {
					const parentPropertyCopy = Object.assign({}, parentProperty);
					const propIndex = properties.findIndex((item, index) => item.key === parentProperty.key);

					parentPropertyCopy.inherited = true;
					parentPropertyCopy.dynamic = false;
					parentPropertyCopy.overridden = false;
					parentPropertyCopy.nearestParent = parentName;
					parentPropertyCopy.nearestParentRoot = parent.root;
					
					if(propIndex === -1) {
						properties.push(parentPropertyCopy);
					} else {
						parentPropertyCopy.overridden = selfProperties.some((item) => item.key === parentPropertyCopy.key);
						if(parentPropertyCopy.overridden) {
							parentPropertyCopy.value = selfProperties[propIndex].value;
							parentPropertyCopy.type = selfProperties[propIndex].type;
						}
						properties[propIndex] = parentPropertyCopy;
					}

					dynamicProperties = dynamicProperties.filter((item) => item.key !== parentProperty.key);
				}
			}

			cls.properties = properties;
			cls.dynamicProperties = dynamicProperties;
			cls.documentableProperties = [...cls.statics, ...cls.properties, ...cls.dynamicProperties].filter(prop => !prop.inherited || prop.overridden).length + 1; // +1 for __self__

			classMap[className] = cls;

			progress && progress.next();
		}
	}

	private static getSuperclassName(z8Cls: any): string | null {
		return z8Cls.superclass.$className || null;
	}

	private static getClassName(z8Cls: any): string {
		return z8Cls.prototype.$className;
	}

	private static getShortClassName(z8Cls: any): string | null {
		const prototype = z8Cls.prototype;
		const superclass = z8Cls.superclass;

		return prototype.$shortClassName !== superclass.$shortClassName ? prototype.$shortClassName : null;
	}

	// TODO: remove previousSuperclassName when Z8.define will be fixed
	private static getParentsBranch(z8Cls: any, Z8Classes: any, classMap: ClassMap, parentsBranch: string[], previousSuperclassName: string | null): string[] | void {
		const clsName = Analyzer.getClassName(z8Cls);
		const superClsName = Analyzer.getSuperclassName(z8Cls);
		
		if(!superClsName || previousSuperclassName === superClsName)
			return parentsBranch.reverse();

		parentsBranch.push(superClsName);

		if(clsName !== superClsName)
			classMap[superClsName] = Analyzer.appendChild(classMap[superClsName], clsName);
		classMap[superClsName] = Analyzer.appendChildren(classMap[superClsName], parentsBranch.filter((parentClsName) => { return parentClsName !== superClsName; }));
		return Analyzer.getParentsBranch(Z8Classes[superClsName], Z8Classes, classMap, parentsBranch, superClsName);
	}

	private static getMixins(z8Cls: any, classMap: ClassMap): string[] {
		if(!z8Cls.prototype.$mixins)
			return [];
		return z8Cls.prototype.$mixins.map((mixin: any) => {
			const mixinClassName = Analyzer.getClassName(mixin);
			Analyzer.appendMixedIn(classMap[Analyzer.getClassName(mixin)], Analyzer.getClassName(z8Cls));
			return mixinClassName;
		});
	}

	private static getProperties(z8Cls: any): Z8ClassProperties[] {
		const prototype = z8Cls.prototype;
		const superclass = z8Cls.superclass;
		const excludeList = ['self', '$class', '$className', '$shortClassName', '$previous', '$owner', 'superclass', 'callParent', 'newInstance'];
		const notInherited = ['callParent', 'newInstance'];

		const superclassProperties = Object.keys(superclass).filter((key) => {
			return !excludeList.includes(key);
		});

		return Object.keys(prototype).filter((key) => {
			return !excludeList.includes(key);
		}).map((key) => {
			const value = prototype[key];
			const type = Analyzer.propertyType(typeof value);
			const strValue = Analyzer.propertyValueToString(key, value, type);
			const overridden = superclassProperties.includes(key) && !notInherited.includes(key);

			return { static: false, type: strValue.indexOf('class__') === 0 ? 'class' : type, key: key, value: strValue.replace('class__', ''), inherited: overridden, dynamic: false, overridden: overridden };
		});
	}

	private static getStatics(z8Cls: any): Z8ClassProperties[] {
		const excludeList = ['self', '$class', '$className', '$shortClassName', '$previous', '$owner', 'superclass'];

		return Object.keys(z8Cls).filter((key) => {
			return !excludeList.includes(key);
		}).map((key) => {
			const value = z8Cls[key];
			const type = Analyzer.propertyType(typeof value);
			try {
				const strValue = Analyzer.propertyValueToString(key, value, type);
				return { static: true, type: strValue.indexOf('class__') === 0 ? 'class' : type, key: key, value: strValue.replace('class__', ''), inherited: false, dynamic: false, overridden: false };
			} catch(e) {
				console.log(Analyzer.getClassName(z8Cls), value, type);
				throw e;
			}
		});
	}

	private static propertyType(type: string): string {
		if(type === 'function')
			return 'method';
		return type;
	}

	private static propertyValueToString(key: string, value: any, type: string): string {
		if(value == null)
			return 'null';

		switch(type) {
		case 'method':
			return value.prototype && value.prototype.$class ? `class__${Analyzer.getShortClassName(value) || Analyzer.getClassName(value)}` : `${key}(${value.toString().split('(')[1].split(')')[0]})`;
		case 'object':
			return value.prototype ? Analyzer.getShortClassName(value) || Analyzer.getClassName(value) : (value.$className ? value.$shortClassName || value.$className : JSON.stringify(value));
		default:
			return value.toString();
		}
	}

	// ClassMapEntity methods
	private static appendChild(cls: ClassMapEntity, childName: string): ClassMapEntity {
		if(!cls.children.includes(childName))
			cls.children.push(childName);
		return cls;
	}

	private static appendChildren(cls: ClassMapEntity, childNames: string[]): ClassMapEntity {
		for(const childName of childNames)
			Analyzer.appendChild(cls, childName);
		return cls;
	}

	private static setParentsBranch(cls: ClassMapEntity, parentsBranch: string[]): ClassMapEntity {
		cls.parentsBranch = parentsBranch;
		return cls;
	}

	private static setExtends(cls: ClassMapEntity, className: string | null): ClassMapEntity {
		cls.extends = className;
		return cls;
	}

	private static setShortName(cls: ClassMapEntity, shortName: string | null): ClassMapEntity {
		cls.shortName = shortName;
		return cls;
	}

	private static setMixins(cls: ClassMapEntity, mixins: string[]): ClassMapEntity {
		cls.mixins = mixins;
		return cls;
	}

	private static appendMixedIn(cls: ClassMapEntity, mixedTo: string): ClassMapEntity {
		if(!cls.mixedIn.includes(mixedTo))
			cls.mixedIn.push(mixedTo);
		return cls;
	}

	private static setStatics(cls: ClassMapEntity, statics: Z8ClassProperties[]): ClassMapEntity {
		cls.statics = statics;
		return cls;
	}

	private static setProperties(cls: ClassMapEntity, properties: Z8ClassProperties[]): ClassMapEntity {
		cls.properties = cls.properties.concat(properties);
		cls.dynamicProperties = cls.dynamicProperties.filter(item1 => !cls.properties.some(item2 => item2.key === item1.key)).filter(item1 => !cls.statics.some(item2 => item2.key === item1.key));
		return cls;
	}
}

export default Analyzer;